from typing import List, Union
from os import environ as env

CREATE_NO_DATA_FLAGS = ['-NO']
CREATE_WITH_DATA_FLAGS = ['-Name', '-Description', '-File']
EDIT_NO_DATA_FLAGS = ['-NO']
EDIT_WITH_DATA_FLAGS = ['-File']
EXECUTE_NO_DATA_FLAGS = ['-NO']
EXECUTE_WITH_DATA_FLAGS = []
GENERAL_FLAG = ['-No-Warning', '-No-Rollback']

def get_flags(args: List[str]) -> List[str]:
    return [_ for _ in args if _.startswith('-')]

def get_parsed_flags(args: List[str]) -> dict[str, Union[str, None]]:
    flags = {}
    argFlags = [_ for _ in args if _.startswith('-')]
    for _ in argFlags:
        if '=' in _:
            f = _.split('=')
            flags[f[0]] = f[1]
        else:
            flags[_] = None
        #Check Flags for global settings
        if _ in GENERAL_FLAG:
            if _ == '-No-Warning':
                env['-No-Warning'] = '1'
            if _ == '-No-Rollback':
                env['-No-Rollback'] = '1'
    return flags

def check_flag_validity(f: dict[str, Union[str, None]], op: str) -> bool:
    """
    Analyses the flag list sent for errors or missing parameters.
    Params:
        `f`: Parsed Flag List;
        `op`: Current Operation being executed;
    """
    operation_wd: List[str] = eval(f"{op.upper()}_WITH_DATA_FLAGS")
    operation_nd: List[str] = eval(f"{op.upper()}_NO_DATA_FLAGS")
    for _ in f.keys():
        if _ in operation_wd and f[_] == None:
            return False
        if _ in operation_nd and f[_] != None:
            f[_] = None
    return True


import os
import shutil
from logging import Logger
from .inspector import implements
from .types import (
    Task, 
    ParserType as Parser, 
    OperationType as Operation
)
from zipfile import ZipFile, ZIP_DEFLATED

def get_file_name(p: str) -> str:
    if '/' in p:
        return p.split('/')[-1]
    return p

@implements(Operation)
class Copy:
    "Copy Action"
    
    __annotations__ = {
        "name": "Copy Action",
        "intent": "Create a duplicate of a specific file on a different location"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'copy'
        self.handle_references()

    def execute(self) -> None:
        files = self.context._get_all_file_paths(self.task['origin']) if self.task['subfolders'] == True else os.listdir(self.task['origin'])
        if self.task['target'] == '*':
            self.__execute(files)
        elif '*' in self.task['target']:
            _files = [_ for _ in files if get_file_name(_).endswith(self.task['target'].split('.')[1])]
            self.__execute(_files)
        else:
            shutil.copyfile(f"{self.task['origin']}/{self.task['target']}", f"{self.task['destination']}/{self.task['target']}")
        #self.logger.debug("Test")

    def rollback(self) -> None:
        for file in self.affected_files:
            file_name = get_file_name(file)
            shutil.copyfile(f"{self.task['destination']}/{file_name}", f"{self.task['origin']}/{file_name}")
        self.logger.warn(f"Rolled back \"{self.task['name']}\" task")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state

    def __execute(self, files: 'list[str]') -> None:
        "Copy files execution action"
        for f in files:
            ori_path = f if self.task['subfolders'] == True else f"{self.task['origin']}/{f}"
            self.affected_files.append(ori_path)
            shutil.copyfile(f"{ori_path}", f"{self.task['destination']}/{get_file_name(f)}")
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

@implements(Operation)
class Move:
    "Move Action"

    __annotations__ = {
        "name": "Move Action",
        "intent": "Move file/files from one location to another"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'move'
        self.handle_references()

    def execute(self) -> None:
        files = self.context._get_all_file_paths(self.task['origin'])
        if self.task['target'] == '*':
            self.__execute(files)
        elif '*' in self.task['target']:
            _files = [_ for _ in files if get_file_name(_).endswith(self.task['target'].split('.')[1])]
            self.__execute(_files)
        else:
            shutil.copyfile(f"{self.task['origin']}/{self.task['target']}", f"{self.task['destination']}/{self.task['target']}")

    def rollback(self) -> None:
        for file in self.affected_files:
            file_name = get_file_name(file)
            shutil.move(f"{self.task['destination']}/{file_name}", f"{self.task['origin']}/{file_name}")
        self.logger.warn(f"Rolled back \"{self.task['name']}\" task")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state
    
    def __execute(self, files: 'list[str]') -> None:
        "Move files execution action"
        for f in files:
            ori_path = f
            self.affected_files.append(ori_path)
            shutil.move(f"{ori_path}", f"{self.task['destination']}/{get_file_name(f)}")
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

@implements(Operation)
class Delete:
    "Delete Action"

    __annotations__ = {
        "name": "Delete Action",
        "intent": "Delete file/files from the system"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'delete'
        self.handle_references()

    def execute(self) -> None:
        fp = []
        if self.task['target'] == '*':
            fp = self.context._get_all_file_paths(self.task['destination']) if self.task['subfolders'] == True else os.listdir(self.task['destination'])
        elif '*' in self.task['target']:
            all_files = self.context._get_all_file_paths(self.task['destination'])
            fp = [_ for _ in all_files if get_file_name(_).lower().endswith(self.task['target'].split('.')[1])]
        for _ in fp:
            os.remove(_)

    def rollback(self) -> None:
        self.logger.warn("No rollback support for Delete Action")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

@implements(Operation)
class Zip:
    "Zip Action"

    __annotations__ = {
        "name": "Zip Action",
        "intent": "Group files into one zipped folder"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'zip'
        self.handle_references()

    def execute(self) -> None:
        fp = []
        files = []
        if self.task['target'] == '*':
            fp = self.context._get_all_file_paths(self.task['destination']) if self.task['subfolders'] == True else [f"{self.task['destination']}/{_}" for _ in os.listdir(self.task['destination'])]
            files = os.listdir(self.task['destination'])
        elif '*' in self.task['target']:
            ending = self.task['target'].replace('*', '')
            fps = self.context._get_all_file_paths(self.task['destination']) if self.task['subfolders'] == True else [f"{self.task['destination']}/{_}" for _ in os.listdir(self.task['destination'])]
            fp = [_ for _ in fps if get_file_name(_).endswith(ending)]
            files = [_ for _ in os.listdir(self.task['destination']) if get_file_name(_).lower().endswith(ending)]

        with ZipFile(f"{self.task['destination']}/{self.task['rename']}.zip", 'w', ZIP_DEFLATED) as zip:
            for i, _ in enumerate(fp):
                if "deflate" in self.task.keys():
                    if self.task['deflate'] == True:
                        zip.write(_, files[i])
                    else:
                        zip.write(_)
                else:
                    zip.write(_)
            zip.close()

    def rollback(self) -> None:
        #Step 1: Extract Zipfile
        #Step 2: Delete created Zip
        with ZipFile(f"{self.task['destination']}/{self.task['rename']}.zip", 'r', ZIP_DEFLATED) as zip:
            zip.extractall(self.task['destination'])
        os.remove(f"{self.task['destination']}/{self.task['rename']}.zip")
        self.logger.warn(f"Rolled back \"{self.task['name']}\" task")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

@implements(Operation)
class Command:
    "Command Action"

    __annotations__ = {
        "name": "Command Action",
        "intent": "Execute CLI commands"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag

    def execute(self) -> None:
        raise NotImplementedError("Command action has not been implemented yet")

    def rollback(self) -> None:
        pass

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state

@implements(Operation)
class Registry:
    "Registry Action"

    __annotations__ = {
        "name": "Registry Action",
        "intent": "Change Windows Registry values/keys"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        if ctx.system != "Windows":
            ctx.abort(f"Operation not available on {ctx.system} Systems!")
        import regutils
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'registry'
        #with sub operations
        #Params-> Key, SubKey, value
        #Must have Rollback
    
    def execute(self) -> None:
        raise NotImplementedError("Registry action has not been implemented yet")
    
    def rollback(self) -> None:
        pass

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state

@implements(Operation)
class Input:
    "Input Action"

    __annotations__ = {
        "name": "Input Action",
        "intent": "Ask User for Input in the console"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.value = None #where the value of the Input will be kept
        self.__internal_state = True #Faulty execution flag
        self._type = 'input'
        self.handle_references()
    
    def execute(self) -> None:
        question = input(self.task["question"])
        self.value = question

    
    def rollback(self) -> None:
        self.logger.warn("No rollback support for Input Action")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

@implements(Operation)
class Echo:
    "Echo Action"

    __annotations__ = {
        "name": "Echo Action",
        "intent": "Print a value to the console"
    }

    def __init__(self, ctx: Parser, task: Task, logger: Logger) -> None:
        self.context = ctx #Parser Context
        self.task = task #Current assigned Task
        self.logger = logger
        self.affected_files: list[str] = []
        self.__internal_state = True #Faulty execution flag
        self._type = 'echo'
        self.handle_references()
    
    def execute(self) -> None:
        value = self.task['value']
        self.logger.debug(f"Output of \"{self.task['name']}\" Task ➡ {value}")

    
    def rollback(self) -> None:
        self.logger.warn("No rollback support for Echo Action")

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        self.__internal_state = state
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        return self.__internal_state
    
    def handle_references(self) -> None:
        for key in self.task.keys():
            if type(self.task[key]) == str and self.task[key].startswith('$'):
                if '.' in self.task[key]:
                    _ = self.task[key].split('.')
                    step = self.context._get_step_reference(self.task, _[0])
                    self.task[key] = step[_[1]]
                else:
                    step = self.context._get_step_reference(self.task, self.task[key])
                    self.task[key] = step[key]

#Add Encrypt Action
#Params-> Algorithm, What


import inspect


def implements(interface_cls):
    def _decorator(cls):
        verify_methods(interface_cls, cls)
        verify_properties(interface_cls, cls)
        verify_attributes(interface_cls, cls)
        return cls

    return _decorator


def verify_methods(interface_cls, cls):
    methods_predicate = lambda m: inspect.isfunction(m) or inspect.ismethod(m)
    for name, method in inspect.getmembers(interface_cls, methods_predicate):
        signature = inspect.signature(method)
        cls_method = getattr(cls, name, None)
        cls_signature = inspect.signature(cls_method) if cls_method else None
        if cls_signature != signature:
            raise NotImplementedError(
                f"'{cls.__name__}' must implement method '{name}({signature})' defined in interface '{interface_cls.__name__}'"
            )


def verify_properties(interface_cls, cls) -> None:
    prop_attrs = dict(
        fget='getter', 
        fset='setter', 
        fdel='deleter'
    )
    for name, prop in inspect.getmembers(interface_cls, inspect.isdatadescriptor):
        cls_prop = getattr(cls, name, None)
        for attr in prop_attrs:
            # instanceof doesn't work for class function comparison
            if type(getattr(prop, attr, None)) != type(getattr(cls_prop, attr, None)):
                raise NotImplementedError(
                    f"'{cls.__name__}' must implement a {prop_attrs[attr]} for property '{name}' defined in interface '{interface_cls.__name__}'"  # flake8: noqa
                )

def verify_attributes(interface_cls, cls) -> None:
    interface_attributes = get_attributes(interface_cls)
    cls_attributes = get_attributes(cls)
    for missing_attr in (interface_attributes - cls_attributes):
        raise NotImplementedError(
            f"'{cls.__name__}' must have class attribute '{missing_attr}' defined in interface '{interface_cls.__name__}'"
        )


def get_attributes(cls) -> set[str]:
    boring = dir(type('dummy', (object,), {}))
    return set(item[0] for item in inspect.getmembers(cls) if item[0] not in boring and not callable(item[1]))


import json
import os
import sys
import os.path as Path
from .operations import *
import platform
from .types import *
from logging import Logger
from typing import Union, List
from webbrowser import open as FileOpener

class Parser:

    def __init__(self, task: str, logger: Logger) -> None:
        self.__first_execution_routine()
        self.task: InstructionSet = json.load(open(f'{self.default_location}/{task}.tasker.json','r'))
        analysis = self.__analyse_keys()
        if not analysis[0]:
            logger.error(f'{analysis[3]} \"{analysis[1]}\" in {analysis[2]}')
            sys.exit(1)
        self.__optional_parameters()
        self.logger = logger
        self.__change_relative_locations(Path.expanduser('~'))
        self.__executed_tasks: List[Task] = []
        self.__operation_stack: list[OperationType] = []
        self.__supported_os = ['Windows'] #List of Tasker supported OSes
    
    def execute(self) -> None:
        self.task['tasks'] = sorted(self.task['tasks'], key=lambda d: d['step'])
        for task in self.task['tasks']:
            if self.__execute(task):
                self.logger.debug(f"Task \"{task['name']}\" - OK")
            else:
                self.logger.error(f"Task \"{task['name']}\" - ERROR")
        #Reverse Operation Stack
        #Do this to use rollback feature on a reverse order
        self.__operation_stack.reverse()
        tick = True
        for operation in self.__operation_stack:
            if not operation.get_state() and '-No-Rollback' not in os.environ:
                if tick:
                    print()
                    print("--------------  Rollbacks  --------------")
                    print()
                    tick = False
                operation.rollback()
    
    def warn_user(self) -> None:
        "Verifies if current OS is one of the allowed ones"
        self.system = platform.system()
        if platform.system() not in self.__supported_os and ('-No-Warning' not in os.environ):
            ans = input(f"'{platform.system()}' is not part of the current supported OS list.\nAre you sure you want to continue? Y/n\n")
            if ans.lower() == 'y':
                pass
            elif ans.lower() == 'n':
                sys.exit(0)
            else:
                self.logger.error("Answer not allowed. Aborting...")
                sys.exit(1)
    
    def abort(self, reason: str) -> None:
        self.logger.error(reason)
        sys.exit(1)

    def __execute(self, task: Task) -> bool:
        try:
            self.__check_destination_path(task, DESTINATION_CHECK_MAP[task['operation']])
            if task['operation'] == 'copy':
                c = Copy(self, task, self.logger)
                self.__operation_stack.append(c)
                c.execute()
            elif task['operation'] == 'move':
                m = Move(self, task, self.logger)
                self.__operation_stack.append(m)
                m.execute()
            elif task['operation'] == 'delete':
                d = Delete(self, task, self.logger)
                self.__operation_stack.append(d)
                d.execute()
            elif task['operation'] == 'zip':
                z = Zip(self, task, self.logger)
                self.__operation_stack.append(z)
                z.execute()
            elif task['operation'] == 'command':
                command = Command(self, task, self.logger)
                self.__operation_stack.append(command)
                command.execute()
            elif task['operation'] == 'input':
                i = Input(self, task, self.logger)
                self.__operation_stack.append(i)
                i.execute()
            elif task['operation'] == 'echo':
                e = Echo(self, task, self.logger)
                self.__operation_stack.append(e)
                e.execute()
            else:
                raise Exception(f"{task['operation']} is an Unknown Operation")
            self.__executed_tasks.append(task)
            return True
        except Exception as e:
            self.__operation_stack[-1].set_state(False)
            return False
    
    def __check_destination_path(self, task: Task, needs_path_check: bool = True) -> None:
        "Check destination path if requested end folder is present. If not, create it."
        if needs_path_check:
            if 'destination' in task.keys():
                if '$' in task['destination']:
                    splitted_path = task['destination'].split("/")[-1]
                    if '.' in splitted_path:
                        _ = splitted_path.split(".")
                        ref = self._get_step_reference(task, _[0])
                        task['destination'] = ref[_[1]]
                    else:
                        ref = self._get_step_reference(task, task['destination'])
                        task['destination'] = ref['destination']
            if 'origin' in task.keys():
                if '$' in task['origin']:
                    splitted_path = task['origin'].split("/")[-1]
                    if '.' in splitted_path:
                        _ = splitted_path.split(".")
                        ref = self._get_step_reference(task, _[0])
                        task['origin'] = ref[_[1]]
                    else:
                        ref = self._get_step_reference(task, task['destination'])
                        task['destination'] = ref['destination']
            if 'destination' not in task.keys():
                if task['target'].startswith('$'):
                    if '.' in task['target']:
                        _ = task['target'].split(".")
                        ref = self._get_step_reference(task, _[0])
                        task['destination'] = ref[_[1]]
                    else:
                        ref = self._get_step_reference(task, task['target'])
                        task['destination'] = ref['destination']
                else:
                    self.logger.debug("Destination parameter is not present")
                    raise Exception()
            destination = task['destination'].split('/')[-1]
            destination_parent = '/'.join(_ for _ in task['destination'].split('/')[:-1])
            if destination not in os.listdir(f"{destination_parent}"):
                os.mkdir(f"{task['destination']}")
    
    def __analyse_keys(self) -> 'tuple[bool, Union[str,None], Union[str,None], Union[str,None]]':
        "Verify if the structure of the Instruction Set is defined correctly"
        for key in OP_INSTRUCTION:
            if key not in self.task.keys():
                return (False, key, 'Definition', 'Missing key')
        if len(self.task['tasks']) > 0:
            for _task in self.task['tasks']:
                for key in OP_TASK:
                    if key not in _task.keys():
                        return (False, key, 'Task', 'Missing key')
                op = _task['operation']
                if op not in OPERATIONS:
                    return (False, op, f"\"{_task['name']}\" Task", 'Unknown Operation')
                operation_keys: list[str] = eval(f'OP_{op.upper()}')
                for key in operation_keys:
                    if key not in _task.keys():
                        #Check for optional parameters
                        if not key.startswith("!"):
                            return (False, key, f"\"{_task['name']}\" Task", 'Missing key')
        return (True, None, None, None)
    
    def __optional_parameters(self) -> None:
        to_change = []
        for i in range(len(self.task['tasks'])):
            #Replace optional parameter identifier with a definitive one
            for key in self.task['tasks'][i].keys():
                if key.startswith("!"):
                    to_change.append([i,key])
        for obj in to_change:
            self.task['tasks'][obj[0]][obj[1].replace("!","")] = self.task['tasks'][obj[0]][obj[1]]
            del self.task['tasks'][obj[0]][obj[1]]

    def _get_all_file_paths(self, directory: str) -> 'List[str]':
        file_paths = []
        for root, _, files in os.walk(directory):
            for filename in files:
                file_paths.append(Path.join(root, filename).replace('\\','/'))
        return file_paths
    
    def __change_relative_locations(self, home: str) -> None:
        for task in self.task['tasks']:
            if (home != None or home != ''):
                if 'destination' in task.keys():
                    task['destination'] = f"{home}/{task['destination']}".replace('\\','/')
                if 'origin' in task.keys() and ":" not in task['origin']:
                    task['origin'] = f"{home}/{task['origin']}".replace('\\','/')
    
    def _get_step_reference(self, task: Task, ref: str, get_from_operation: bool = False) -> Union[Task, dict]:
        #get step in reference
        step_index = next((index for (index, d) in enumerate(self.__executed_tasks) if d['step'] == int(ref.replace('$',''))), None)
        if step_index == None:
            self.logger.error(f"Reference in Task \"{task['name']}\" is either not been executed or doesn't exist.")
            raise Exception()
        vars = {}
        if get_from_operation:
            pass
        for _ in self.__operation_stack:
            if _.task["step"] == step_index:
                vars = _.__dict__
                break
        return dict(self.__executed_tasks[step_index], **vars)
    
    def __first_execution_routine(self) -> None:
        "Create the initial configuration and setup necessary directories"
        def create_initial_config(p: str) -> None:
            with open(f"{p}/.tasker/config.json",'w') as config:
                json.dump({
                    "current_location": p,
                    "default_location": p
                }, config, indent=4)
                config.close()
        root_path = Path.expanduser('~')
        root_folders = os.listdir(root_path)
        if '.tasker' not in root_folders:
            os.mkdir(f"{root_path}/.tasker")
            os.mkdir(f"{root_path}/.tasker/Tasks")
            create_initial_config(root_path)
        else:
            tasker_folder = os.listdir(f"{Path.expanduser('~')}/.tasker")
            if 'config.json' not in tasker_folder:
                create_initial_config(root_path)
        self.default_location = f"{root_path}/.tasker/Tasks"

    # Static Methods

    @staticmethod
    def do_config() -> None:
        "Create the initial configuration and setup necessary directories"
        def create_initial_config(p: str) -> None:
            with open(f"{p}/.tasker/config.json",'w') as config:
                json.dump({
                    "current_location": p,
                    "default_location": p
                }, config, indent=4)
                config.close()
        root_path = Path.expanduser('~')
        root_folders = os.listdir(root_path)
        if '.tasker' not in root_folders:
            os.mkdir(f"{root_path}/.tasker")
            os.mkdir(f"{root_path}/.tasker/Tasks")
            create_initial_config(root_path)
        else:
            tasker_folder = os.listdir(f"{Path.expanduser('~')}/.tasker")
            if 'config.json' not in tasker_folder:
                create_initial_config(root_path)
    
    @staticmethod
    def list_all_tasks() -> List[str]:
        "Lists all Task templates created"
        Parser.do_config()
        return [_.replace('.tasker.json', '') for _ in os.listdir(f"{Path.expanduser('~')}/.tasker/Tasks")]
    
    @staticmethod
    def create_new_task(file_name:str, name: str, description: str) -> InstructionSet:
        Parser.do_config()
        i: InstructionSet = {
            'name': name,
            'description': description,
            'tasks': [
                
            ]
        }
        with open(f"{Path.expanduser('~')}/.tasker/Tasks/{file_name}.tasker.json", 'w') as instruction_set:
            json.dump(i, instruction_set, indent=4)
            instruction_set.close()
        return i
    
    @staticmethod
    def open_file_for_edit(file: str) -> None:
        Parser.do_config()
        FileOpener(f"{Path.expanduser('~')}/.tasker/Tasks/{file}.tasker.json")


"Windows Registry Utils file"
import winreg
#import os.path as Path
from os import listdir, remove

ROOTS: dict[str,int] = {
    "classes-root": winreg.HKEY_CLASSES_ROOT,
    "current-user": winreg.HKEY_CURRENT_USER,
    "current-config": winreg.HKEY_CURRENT_CONFIG,
    "local-machine": winreg.HKEY_LOCAL_MACHINE,
    "users": winreg.HKEY_USERS,
}

def parse_input(k: str):
    if k == '':
        raise Exception("Cannot provide empty Registry Path")
    str_list = k.split('>')
    root: int = ROOTS[str_list[0]]
    str_list.pop(0)
    return [root, [_ for _ in str_list if _ != '']]

""" def get_all_keys(root: str) -> List[str]:
    keys = []
    i = 0
    try:
        _ = eval(f"winreg.HKEY_{root.upper()}")
        while True:
            k = winreg.EnumKey(_, i)
            keys.append(k)
            i += 1
    except Exception as e:
        pass
    return keys """

def create_key(path: str, key_name: str) -> None:
    k = parse_input(path)
    with winreg.OpenKey(k[0], r'\\'.join(k[1]), reserved=0, access=winreg.KEY_CREATE_SUB_KEY) as key:
        winreg.CreateKeyEx(key, key_name)

def set_key_value(path: str, name: str, _type: int, value: str) -> None:
    k = parse_input(path)
    with winreg.OpenKey(k[0], r'\\'.join(k[1]), reserved=0, access=winreg.KEY_SET_VALUE) as key:
        winreg.SetValueEx(key, name, 0, _type, value)

def get_key_value(path: str) -> str:
    k = parse_input(path)
    v = k[1][-1]
    k[1].pop(-1)
    with winreg.OpenKey(k[0], r'\\'.join(k[1]), access=winreg.KEY_READ) as key:
        return winreg.QueryValueEx(key, rf'{v}')[0]

def backup(root: str, path: str, fname: str):
    files = listdir(path)
    if f"{fname}.reg" in files:
        remove(f"{path}/{fname}.reg")
    r = winreg.ConnectRegistry(None, ROOTS[root])
    with winreg.OpenKey(r, '', reserved=0, access=winreg.KEY_ALL_ACCESS) as key:
        winreg.SaveKey(key, f"{path}/{fname}.reg")


if __name__ == '__main__':
    #print(get_all_keys('LOCAL_MACHINE'))
    #print(get_key_value("local-machine>SOFTWARE>Autodesk>3dsMax>Current Version>Executable"))
    print(get_key_value("local-machine>SOFTWARE>BlueStacksInstaller>MachineID"))
    #set_key_value("local-machine>SOFTWARE>BlueStacksInstaller", "MachineIDs", winreg.REG_SZ, 'a')
    #print(get_key_value("local-machine>SOFTWARE>BlueStacksInstaller>MachineID"))
    #get_key_value("classes-root>.iso")
    #backup("current-config", f"{Path.expanduser('~')}/Desktop/", 'current_config')
    #create_key("local-machine>SOFTWARE>BlueStacksInstaller", "MachineIDs")
    #print(get_key_value("local-machine>SOFTWARE>ATI Technologies>Install>Packages>W-06-3U01-000-001-044-001-00-25>breboot_req"))


from typing import TypedDict, List, Literal, Dict
from logging import Logger

#Structure definitions
OP_INSTRUCTION = ['name', 'description', 'tasks']
OP_TASK = ['name', 'step', 'operation']

#Operation Key values
OP_COPY = ['target', 'origin', 'destination', 'subfolders']
OP_MOVE = ['target', 'origin', 'destination']
OP_COMMAND = ['target', 'origin', 'destination']
OP_DELETE = ['target', '!destination']
OP_ZIP = ['target', 'rename', '!deflate', '!destination']
OP_INPUT = ['question']
OP_ECHO = ['value']

#Available Operations
OPERATIONS = ['copy', 'zip', 'move', 'delete', 'command', 'input', 'echo']
LIST_OPERATIONS = Literal['copy', 'zip', 'move', 'delete', 'input', 'echo']

#Structure Definition for task
class Task(TypedDict):
    name: str
    step: int
    operation: str
    target: str
    origin: str
    destination: str
    rename: str
    subfolders: bool
    deflate: bool
    question: str
    value: str

#Structure Definition for instruction_set
class InstructionSet(TypedDict):
    name: str
    description: str
    tasks: List[Task]

class ParserType:

    system: str
    task: InstructionSet
    logger: Logger

    def execute(self, task: str, logger: Logger, default_location: str) -> None:
        pass
    
    def warn_user(self) -> None:
        pass

    def abort(self, reason: str) -> None:
        pass

    def __execute(self, task: Task) -> bool:
        pass

    def __check_destination_path(self, task: Task) -> None:
        pass

    def __analyse_keys(self) -> 'tuple[bool, str, str, str]':
        pass

    def __optional_parameters(self) -> None:
        pass

    def _get_all_file_paths(self, directory: str) -> 'List[str]':
        pass

    def _get_file_name(self, p: str) -> str:
        pass

    def __change_relative_locations(self, home: str) -> None:
        pass

    def _get_step_reference(self, task: Task, ref: str, get_from_operation: bool = False) -> Task:
        pass

class OperationType:
    "Skeleton class for creating other, more complex Operations"

    __internal_state: bool
    task: Task
    context: ParserType
    logger: Logger
    _type: str

    def execute(self) -> None:
        pass
    
    def rollback(self) -> None:
        pass
    
    def get_state(self) -> bool:
        "Returns the of the Internal Fault flag"
        pass

    def set_state(self, state: bool) -> None:
        "Sets the state for the Internal Fault flag"
        pass

DESTINATION_CHECK_MAP: Dict[str, bool] = {
    "copy": True,
    "delete": True,
    "echo": False,
    "input": False,
    "move": True,
    "registry": False,
    "zip": True
}


import os
import sys

from setuptools import setup, find_packages

if sys.argv[-1] == 'build':
    os.system('python setup.py sdist')
    sys.exit()

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist')
    os.system('twine upload dist/*')
    sys.exit()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    author = "Carlos Silva",
    author_email = "carlos.miguel.silva@protonmail.com",
    name = "pyTasker",
    description = "Run pipelines on your own computer for better automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version = "0.2.1",
    url="https://github.com/carlossilva2/pyTasker",
    packages = find_packages(),
    license="GPLv3",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Natural Language :: English',
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development',
        'Typing :: Typed'
    ],
    python_requires='>=3.9',
    entry_points = {
        "console_scripts": [
            "tasker = Tasker.__main__:main"
        ]
    },
    install_requires = [

    ],
    package_data={
        "static": ["*"]
    },
    project_urls={
        "Source": "https://github.com/carlossilva2/pyTasker"
    }
)


import sys
import logging

from .parser import Parser
from .cli import get_parsed_flags, check_flag_validity

__version__ = '0.2.1'

HELP_TEXT = """
Usage:
  tasker <command>
Commands:
  list                      Lists all available Tasks in the system.
  execute <task>            Executes the specified Task. Options [-NO].
  create                    Creates a new Task. Options [-NO | -Description | -Name | -File].
  edit                      Opens the specified Task file on the default text editor. Options [-NO | -File]
  help                      Show help for commands.
General Options:
  -v, -version              Show version and exit.
  -h, -help                 Show help.
  -NO                       Makes the parser not output anything to the console.
  -Name                     Task Human Readable Name flag.
  -File                     Task File Name flag.
  -Description              Description for Task template.
  -No-Warning               Removes the verification of the Users OS.
  -No-Rollback              Prevents Tasker to perform Rollback in case of Task failure.
"""
def main() -> None: 
    args = sys.argv[1:]
    logging.basicConfig(level=logging.DEBUG, format='[%(levelname)s][%(asctime)s] → %(message)s',datefmt='%H:%M:%S')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    if len(args) > 0:
        flags = get_parsed_flags(args)
        if args[0] == 'list':
            for task in Parser.list_all_tasks():
                print(f"   ➡ {task}")
        elif args[0] == 'execute' and len(args) >= 2:
            P = Parser(args[1], logger)
            P.warn_user()
            P.execute()
        elif args[0] == 'edit' and len(args) >= 2:
            if not check_flag_validity(flags, 'edit'):
                logger.error('Flags were not constructed properly')
                sys.exit(1)
            Parser.open_file_for_edit(flags['-File'])
        elif args[0] == 'create':
            if not check_flag_validity(flags, 'create'):
                logger.error('Flags were not constructed properly')
                sys.exit(1)
            Parser.create_new_task(flags['-File'], flags['-Name'], flags['-Description'])
        elif args[0] == '-version' or args[0] == '-v':
            print(f"Tasker V{__version__}")
        elif args[0] == '-help' or args[0] == '-h' or args[0] == 'help':
            print(HELP_TEXT)
        else:
            logger.error('Check Help for command syntax')
    else:
        print(HELP_TEXT)
    sys.exit(1)

if __name__ == '__main__':
    main()